///////////////////////////////////////////////
// Copyright
///////////////////////////////////////////////
//
// FliFlc
// Copyright (c) 2001 Jari Komppa
//
//
///////////////////////////////////////////////
// License
///////////////////////////////////////////////
// 
//     This software is provided 'as-is', without any express or implied
//     warranty.    In no event will the authors be held liable for any damages
//     arising from the use of this software.
// 
//     Permission is granted to anyone to use this software for any purpose,
//     including commercial applications, and to alter it and redistribute it
//     freely, subject to the following restrictions:
// 
//     1. The origin of this software must not be misrepresented; you must not
//        claim that you wrote the original software. If you use this software
//        in a product, an acknowledgment in the product documentation would be
//        appreciated but is not required.
//     2. Altered source versions must be plainly marked as such, and must not be
//        misrepresented as being the original software.
//     3. This notice may not be removed or altered from any source distribution.
// 
// (eg. same as ZLIB license)
// 
//
///////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "fliflc.h"

#pragma pack(1) // Pack structure members to 1 byte boundaries 
typedef struct { // fli header 
    int       fli_size; // Should be same as file size. 
    short int fli_magic; // 0AF12h 
    short int fli_frames; // Frames in flic, 4000 max 
    short int width; // x size 
    short int height;// y size 
    short int depth; // bits per pixel, always 8 in fli/flc 
    short int flags; // Bitmapped flags. 0=ring frame, 1=header updated 
    short int speed; // Delay between frames in ms (or retraces :)) 
    short int reserved1;
    int       created; // MS-dos date of creation 
    int       creator; // SerNo of animator, 0464c4942h for FlicLib 
    int       updated; // MS-dos date of last modify 
    int       updater; // ..
    short int aspectx; // x-axis aspect ratio (320x200: 6)
    short int aspecty; // y-axis aspect ratio (320x200: 5)
    char      reserved2[38];
    int       oframe1; // offset to frame 1 
    int       oframe2; // offset to frame 2 - for looping, jump here 
    char      reserved3[42];
} FLIHEADER;

typedef struct { // frame header 
    int       framesize;
    unsigned short int magic;
    short int chunks;
    unsigned char reserved[8];
} FRAMEHEADER;

typedef struct { // chunk header 
    int       size;
    short int type;
} CHUNKHEADER;

#pragma pack()

/************************************************************************
**  FLI/FLC decoder functions. Part 1: chunk decoders                  **
*************************************************************************/

// "Set whole frame to index 0" -chunk 
void FliAnimation::fli_black() 
{
  memset(mFramebuffer,0,mXSize*mYSize);
}

// Uncompressed frame (extremely rare) 
void FliAnimation::fli_copy(unsigned char *data) 
{
  memcpy(mFramebuffer,data,mXSize*mYSize);
}

// 64 - level palette chunk 
void FliAnimation::fli_colour(unsigned char *cdata) 
{
short int *pktaddress;
unsigned char skip;
unsigned char set;
short int numberpk;
short int packetcount;
int a;
  pktaddress=(short int *)cdata;
  cdata+=2;
  numberpk=*pktaddress;
  for (packetcount=0;packetcount<numberpk;packetcount++) {
    skip=*cdata;
    cdata++;
    set=*cdata;
    cdata++;
    if (set==0) {
      for (a=0;a<768;a++)
        *(mPalette+a)=*(cdata+a);
      mPaletteChange=1;
    } else {
      for (a=0;a<set*3;a++)
        *(mPalette+skip*3+a)=*(cdata+a);
      mPaletteChange=1;
      cdata+=(set*3);
    }
  }
}

// 256 - level palette chunk 
void FliAnimation::fli_colour256(unsigned char *cdata) 
{
short int *pktaddress;
unsigned char skip;
int set;
short int numberpk;
short int packetcount;
int a;
  pktaddress=(short int *)cdata;
  cdata+=2;
  numberpk=*pktaddress;
  for (packetcount=0;packetcount<numberpk;packetcount++) {
    skip=*cdata;
    cdata++;
    set=*cdata;
    cdata++;
    if (set==0) {
      for (a=0;a<768;a++)
        *(mPalette+a)=*(cdata+a)/4;
      mPaletteChange=1;
    } else {
      for (a=0;a<(set*3);a++)
        *(mPalette+a+skip*3)=*(cdata+a)/4;
      mPaletteChange=1;
      cdata+=(set*3);
    }
  }
}

// 8b-based delta 
void FliAnimation::fli_rc(unsigned char*lcdata) 
{
short int *addlines;
short int numlines;
unsigned char *vbuffptr;
short int linecount;
unsigned char pktcount,skip,numpkt,sizecount,databyte;
signed char size;
unsigned char *linestart;
  vbuffptr=mFramebuffer;
  addlines=(short int *) lcdata;
  numlines=*addlines;
  lcdata+=4;
  addlines+=1;
  vbuffptr+=numlines*mXSize;
  numlines=*addlines;
  linestart=vbuffptr;
  for (linecount=0;linecount<numlines;linecount++) {
    vbuffptr=linestart;
    numpkt=*lcdata;
    lcdata++;
    for (pktcount=0;pktcount<numpkt;pktcount++)  {
      skip=*lcdata;
      lcdata++;
      vbuffptr+=skip;
      size=(signed char)*lcdata;
      lcdata++;
      if (size>=0) {
        for (sizecount=0;sizecount<size;sizecount++) {
          *vbuffptr=*lcdata;
          vbuffptr++;
          lcdata++;
        }
      } else {
        size=-size;
        databyte=*lcdata;
        lcdata++;
        for (sizecount=0;sizecount<size;sizecount++) {
          *vbuffptr=databyte;
          vbuffptr++;
        }
      }
    }
  linestart+=mXSize;
  }
}

// 16b-based delta (FLC only) 
void FliAnimation::fli_ss2(unsigned char*lcdata) 
{
short int numlines;
unsigned char *vbuffptr;
short int linecount;
char skip;
int pktcount,sizecount,databyte;
short int numpkt;
signed char size;
unsigned char *linestart;
  vbuffptr=mFramebuffer;
  numlines=*(short int *)lcdata;
  lcdata+=2;
  linestart=vbuffptr;
  for (linecount=0;linecount<numlines;linecount++) {
    vbuffptr=linestart;
    numpkt=*(short int *)lcdata;
    lcdata+=2;
    if (numpkt<=0) {
      numpkt=-numpkt;
      linecount--;
      linestart+=(numpkt-1)*mXSize;
    } else
      for (pktcount=0;pktcount<numpkt;pktcount++)  {
        skip=*lcdata;
        lcdata++;
        vbuffptr+=skip;
        size=*lcdata;
        lcdata++;
        if (size>=0) {
          for (sizecount=0;sizecount<size;sizecount++) {
            *vbuffptr=*lcdata;
            vbuffptr++;
            lcdata++;
            *vbuffptr=*lcdata;
            vbuffptr++;
            lcdata++;
          }
        } else {
          size=-size;
          databyte=*lcdata;
          lcdata++;
          for (sizecount=0;sizecount<size;sizecount++) {
            *vbuffptr=databyte;
            vbuffptr++;
            *vbuffptr=*lcdata;
            vbuffptr++;
          }
        lcdata++;
        }
      }
  linestart+=mXSize;
  }
}

// RLE full frame 
void FliAnimation::fli_brun(unsigned char*brundata) 
{
short int numlines;
unsigned char *vbuffptr;
unsigned char pktcount,numpkt,sizecount;
signed char size;
  vbuffptr=mFramebuffer;
  for (numlines=0;numlines<mYSize;numlines++) {
    numpkt=*brundata;
    brundata++;
    for (pktcount=0;pktcount<numpkt;pktcount++) {
      size=(signed char)*brundata;
      brundata++;
      if (size>=0) {
        for (sizecount=0;sizecount<size;sizecount++) {
          *vbuffptr=*brundata;
          vbuffptr++;
        }
        brundata++;
      } else {
        size=-size;
        for (sizecount=0;sizecount<size;sizecount++) {
          *vbuffptr=*brundata;
          vbuffptr++;
          brundata++;
        }
      }
    }
  }
}

/************************************************************************
**  FLI/FLC decoder functions. Part 2: frame decoder                   **
*************************************************************************/

unsigned char * FliAnimation::decode_chunk(unsigned char *thischunk) 
{
CHUNKHEADER *chunkhead;
unsigned char *nextchunk;
unsigned char *chunkdata;
  chunkhead=(CHUNKHEADER *)thischunk;
  nextchunk=thischunk+chunkhead->size;
  chunkdata=thischunk+6;
  switch (chunkhead->type) {
    case  4: /* 256-level palette (FLC only) */
             fli_colour256(chunkdata);
             break;
    case  7: /* 16b-based delta (FLC only) */
             fli_ss2(chunkdata);
             break;
    case 11: /* 64-level palette */
             fli_colour(chunkdata);
             break;
    case 12: /* 8b-based delta */
             fli_rc(chunkdata);
             break;
    case 13: /* full black frame */
             fli_black();
             break;
    case 15: /* RLE full frame */
             fli_brun(chunkdata);
             break;
    case 16: /* full frame, no compression */
             fli_copy(chunkdata);
             break;
    case 18: /* postage stamp sized image - out of this routine's scope */
             break;
    default: /* unknown/irrelevant */
             break;
  }
  return(nextchunk);
}

unsigned char * FliAnimation::fli_frame() 
{
FRAMEHEADER *thisframe;
unsigned char *nextframe;
short int numchunks;
unsigned char *thischunk;
  thisframe=(FRAMEHEADER *) mNextframe;
  nextframe=mNextframe;
  nextframe+=thisframe->framesize;
  thischunk=(mNextframe+sizeof(FRAMEHEADER));
  if (thisframe->magic==0xf1fa)
  for (numchunks=1;numchunks<=thisframe->chunks;numchunks++) {
    thischunk=decode_chunk(thischunk);
  }
  return(nextframe);
}

/************************************************************************
**  FLI/FLC decoder functions. Part 3: "high level" functions          **
*************************************************************************/

FliAnimation::FliAnimation(char * filename,int offset) 
{
FILE * fli_file;
FLIHEADER header;
  fli_file=fopen(filename,"rb");
  if (fli_file==0) {
    return;
  }
  fseek(fli_file,offset,SEEK_SET);
  fread(&header,1,sizeof(header),fli_file);
  mFramebuffer=new unsigned char[header.width*header.height];
  mXSize=header.width;
  mYSize=header.height;
  mPalette=new unsigned char[768];
  mPaletteChange=0;
  mCurframe=0;
  mLooped=0;
  mMaxframe=header.fli_frames;
  mFlicdata=new unsigned char[header.fli_size-sizeof(header)];
  fread(mFlicdata,1,header.fli_size-sizeof(header),fli_file);
  mNextframe=mFlicdata;
  mLoopframe=mNextframe;
  fclose(fli_file);
  memset(mFramebuffer,0,header.width*header.height);
}

FliAnimation::~FliAnimation() {
    delete[] mFramebuffer;
    delete[] mPalette;
    delete[] mFlicdata;
}

void FliAnimation::renderframe() 
{
  mNextframe=fli_frame();
  mCurframe++;
  if (mCurframe==mMaxframe) {
    mCurframe=0;
    mNextframe=mLoopframe;
    mLooped=1;
  }
}
