///////////////////////////////////////////////
// Copyright
///////////////////////////////////////////////
//
// TextFX6
// Copyright (c) 1995-2009 Jari Komppa
//
//
///////////////////////////////////////////////
// License
///////////////////////////////////////////////
// 
//     This software is provided 'as-is', without any express or implied
//     warranty.    In no event will the authors be held liable for any damages
//     arising from the use of this software.
// 
//     Permission is granted to anyone to use this software for any purpose,
//     including commercial applications, and to alter it and redistribute it
//     freely, subject to the following restrictions:
// 
//     1. The origin of this software must not be misrepresented; you must not
//        claim that you wrote the original software. If you use this software
//        in a product, an acknowledgment in the product documentation would be
//        appreciated but is not required.
//     2. Altered source versions must be plainly marked as such, and must not be
//        misrepresented as being the original software.
//     3. This notice may not be removed or altered from any source distribution.
// 
// (eg. same as ZLIB license)
// 
//
///////////////////////////////////////////////

#include <stdlib.h>
#include <assert.h>
#include "textfx.h"


///////////////////////////////////////////////
// Global data & variables
//

#include "tfx_fontdata.h"

///////////////////////////////////////////////
// Functions
//

TFX_AsciiArt2::TFX_AsciiArt2()
{
    mAsciiMap=NULL;
}

TFX_AsciiArt2::~TFX_AsciiArt2()
{
    delete[] mAsciiMap;
}


#define IS_BIT(a,b) (((a)&(1<<(7-(b))))?1:0)


// This is _old_ code, and I don't feel like rewriting it at the moment.

void TFX_AsciiArt2::BuildLUT()
{
    int startchar=0;
    int endchar=255;
    startchar=32;
    endchar=126;
    int a,b,c,d,e,f,g,h,i;  
    int * reg;
    reg = new int[256*4]; // allocate memory for font analyse data 
    unsigned char * pchar = (unsigned char *)TFX_AsciiFontdata;    
    if (mAsciiMap != NULL)       // regenerating? 
        delete[] mAsciiMap;
    mAsciiMap = new short[16*16*16*16]; // allocate the look-up table 
    // Step one: analyse font

    int chars;
    for (chars = startchar; chars <= endchar; chars++)
    {
        int charoffset = chars * 12;
        int i, j, c;
        int scratch[8*12];
        for (i = 0, c = 0; i < 12; i++)
        {
            for (j = 0; j < 8; j++, c++)
            {
                int cc = IS_BIT(*(pchar+i+charoffset),j) ? 255 : 0;
                scratch[i*8+j] = cc;
            }
        }

        int blurred[12*8];

        for (i = 0, c = 0; i < 12; i++)
        {
            for (j = 0; j < 8; j++, c++)
            {
                blurred[c] =
                    (
                    scratch[((i+0)%12)*8+((j+0)%8)]*4 +
                    scratch[((i+1)%12)*8+((j+0)%8)]*2*((i==11)?0:1) +
                    scratch[((i+11)%12)*8+((j+0)%8)]*2*((i==0)?0:1) +
                    scratch[((i+0)%12)*8+((j+1)%8)]*2*((j==7)?0:1) +
                    scratch[((i+1)%12)*8+((j+1)%8)]*((i==11 || j==7)?0:1) +
                    scratch[((i+11)%12)*8+((j+1)%8)]*((i==0 || j==7)?0:1) +
                    scratch[((i+0)%12)*8+((j+7)%8)]*2*((j==0)?0:1) +
                    scratch[((i+1)%12)*8+((j+7)%8)]*((i==11 || j==0)?0:1) +
                    scratch[((i+11)%12)*8+((j+7)%8)]*((i==0 || j==0)?0:1)
                    ) / 8;
            }
        }

        e=0;
        for (b=0;b<6;b++)
            for (c=0;c<4;c++)
                e+=blurred[b*8+c];
        *(reg+chars*4+0)=e/24;
        e=0;
        for (b=0;b<6;b++)
            for (c=0;c<4;c++)
                e+=blurred[b*8+c+4];
        *(reg+chars*4+1)=e/24;
        e=0;
        for (b=0;b<6;b++)
            for (c=0;c<4;c++)
                e+=blurred[b*8+c+6*8];
        *(reg+chars*4+2)=e/24;
        e=0;
        for (b=0;b<6;b++)
            for (c=0;c<4;c++)
                e+=blurred[b*8+c+6*8+4];
        *(reg+chars*4+3)=e/24;
    }

    // Step two: build lookup table with the good old nearest-value mess 
    for (a=0;a<16;a++) 
    {
        for (b=0;b<16;b++)
        {
            for (c=0;c<16;c++)
            {
                for (d=0;d<16;d++) 
                {
                    f=0xffffff;
                    g=0;
                    for (h=startchar;h<(endchar+1);h++) 
                    {
#define POT(x) ((x)*(x))
                        i=POT(*(reg+h*4+0)-a*16)+
                          POT(*(reg+h*4+1)-b*16)+
                          POT(*(reg+h*4+2)-c*16)+
                          POT(*(reg+h*4+3)-d*16);
                        if (i<f) 
                        {
                            f=i;
                            g=h;
                        }
                    }
                    *(mAsciiMap+(a<<12)+(b<<8)+(c<<4)+d)=0x0700+g;
                }
            }            
        }
    }
    delete[] reg;   
}


void TFX_AsciiArt2::Dump1x(int *aSource, TFXQuad &aSrcQuad, int aSrcPitch, int aTgtX0, int aTgtY0, short *aTarget)
{
    assert(mAsciiMap!=NULL);
    int xsize = (aSrcQuad.x1 - aSrcQuad.x0);
    int ysize = (aSrcQuad.y1 - aSrcQuad.y0);
    int sourceyofs = aSrcQuad.y0 * aSrcPitch;

    for (int y = 0; y < ysize; y++, sourceyofs += aSrcPitch)
    {
        for (int x = 0, targetpos=(aTgtY0 + y) * TFX_ConsoleWidth + aTgtX0, sourcepos = sourceyofs + aSrcQuad.x0; x < xsize; x++, targetpos++, sourcepos++)
        {
            int c1 = aSource[sourcepos];

            int c1a = (((c1 >>  0) & 0xff) * 1 + 
                       ((c1 >>  8) & 0xff) * 6 + 
                       ((c1 >> 16) & 0xff) * 3) / 10;

            int c=((c1a>>4)<<12)+
                  ((c1a>>4)<<8)+
                  ((c1a>>4)<<4)+
                  ((c1a>>4));

            aTarget[targetpos] = mAsciiMap[c];                                 
                                 
        }
    }
}


void TFX_AsciiArt2::Dump2x(int *aSource, TFXQuad &aSrcQuad, int aSrcPitch, int aTgtX0, int aTgtY0, short *aTarget)
{
    assert(mAsciiMap!=NULL);
    int xsize = (aSrcQuad.x1 - aSrcQuad.x0) / 2;
    int ysize = (aSrcQuad.y1 - aSrcQuad.y0) / 2;
    int sourceyofs = aSrcQuad.y0 * aSrcPitch;

    for (int y = 0; y < ysize; y++, sourceyofs += aSrcPitch * 2)
    {
        for (int x = 0, targetpos=(aTgtY0 + y) * TFX_ConsoleWidth + aTgtX0, sourcepos = sourceyofs + aSrcQuad.x0; x < xsize; x++, targetpos++, sourcepos += 2)
        {
            int c1 = aSource[sourcepos];
            int c2 = aSource[sourcepos + 1];
            int c3 = aSource[sourcepos + aSrcPitch];
            int c4 = aSource[sourcepos + aSrcPitch + 1];

            int c1a = (((c1 >>  0) & 0xff) * 1 + 
                       ((c1 >>  8) & 0xff) * 6 + 
                       ((c1 >> 16) & 0xff) * 3) / 10;

            int c2a = (((c2 >>  0) & 0xff) * 1 + 
                       ((c2 >>  8) & 0xff) * 6 + 
                       ((c2 >> 16) & 0xff) * 3) / 10;

            int c3a = (((c3 >>  0) & 0xff) * 1 + 
                       ((c3 >>  8) & 0xff) * 6 + 
                       ((c3 >> 16) & 0xff) * 3) / 10;

            int c4a = (((c4 >>  0) & 0xff) * 1 + 
                       ((c4 >>  8) & 0xff) * 6 + 
                       ((c4 >> 16) & 0xff) * 3) / 10;

            int c=(((c1a>>4)&0xf)<<12)+
                  (((c2a>>4)&0xf)<<8)+
                  (((c3a>>4)&0xf)<<4)+
                  (((c4a>>4)&0xf));

            aTarget[targetpos] = mAsciiMap[c];                                 
                                 
        }
    }
}

void TFX_AsciiArt2::Dump4x(int *aSource, TFXQuad &aSrcQuad, int aSrcPitch, int aTgtX0, int aTgtY0, short *aTarget)
{
    assert(mAsciiMap!=NULL);
    int xsize = (aSrcQuad.x1 - aSrcQuad.x0) / 4;
    int ysize = (aSrcQuad.y1 - aSrcQuad.y0) / 4;
    int sourceyofs = aSrcQuad.y0 * aSrcPitch;

    for (int y = 0; y < ysize; y++, sourceyofs += aSrcPitch * 4)
    {
        for (int x = 0, targetpos=(aTgtY0 + y) * TFX_ConsoleWidth + aTgtX0, sourcepos = sourceyofs + aSrcQuad.x0; x < xsize; x++, targetpos++, sourcepos += 4)
        {
            int rs, gs, bs;

            rs = 0; gs = 0; bs = 0;

            int i, j;

            for (i = 0; i < 2; i++)
            {
                for (j = 0; j < 2; j++)
                {
                    bs += (aSource[sourcepos + j + i * aSrcPitch] >> 0) & 0xff;
                    gs += (aSource[sourcepos + j + i * aSrcPitch] >> 8) & 0xff;
                    rs += (aSource[sourcepos + j + i * aSrcPitch] >> 16) & 0xff;
                }
            }
            int c1 = (((bs >> 2) & 0xff) << 0) |
                     (((gs >> 2) & 0xff) << 8) |
                     (((rs >> 2) & 0xff) << 16);

            rs = 0; gs = 0; bs = 0;

            for (i = 0; i < 2; i++)
            {
                for (j = 0; j < 2; j++)
                {
                    bs += (aSource[2 + sourcepos + j + i * aSrcPitch] >> 0) & 0xff;
                    gs += (aSource[2 + sourcepos + j + i * aSrcPitch] >> 8) & 0xff;
                    rs += (aSource[2 + sourcepos + j + i * aSrcPitch] >> 16) & 0xff;
                }
            }
            int c2 = (((bs >> 2) & 0xff) << 0) |
                     (((gs >> 2) & 0xff) << 8) |
                     (((rs >> 2) & 0xff) << 16);

            rs = 0; gs = 0; bs = 0;

            for (i = 0; i < 2; i++)
            {
                for (j = 0; j < 2; j++)
                {
                    bs += (aSource[aSrcPitch + sourcepos + j + i * aSrcPitch] >> 0) & 0xff;
                    gs += (aSource[aSrcPitch + sourcepos + j + i * aSrcPitch] >> 8) & 0xff;
                    rs += (aSource[aSrcPitch + sourcepos + j + i * aSrcPitch] >> 16) & 0xff;
                }
            }
            int c3 = (((bs >> 2) & 0xff) << 0) |
                     (((gs >> 2) & 0xff) << 8) |
                     (((rs >> 2) & 0xff) << 16);

            rs = 0; gs = 0; bs = 0;

            for (i = 0; i < 2; i++)
            {
                for (j = 0; j < 2; j++)
                {
                    bs += (aSource[2 + aSrcPitch + sourcepos + j + i * aSrcPitch] >> 0) & 0xff;
                    gs += (aSource[2 + aSrcPitch + sourcepos + j + i * aSrcPitch] >> 8) & 0xff;
                    rs += (aSource[2 + aSrcPitch + sourcepos + j + i * aSrcPitch] >> 16) & 0xff;
                }
            }
            int c4 = (((bs >> 2) & 0xff) << 0) |
                     (((gs >> 2) & 0xff) << 8) |
                     (((rs >> 2) & 0xff) << 16);

            int c1a = (((c1 >>  0) & 0xff) * 1 + 
                       ((c1 >>  8) & 0xff) * 6 + 
                       ((c1 >> 16) & 0xff) * 3) / 10;

            int c2a = (((c2 >>  0) & 0xff) * 1 + 
                       ((c2 >>  8) & 0xff) * 6 + 
                       ((c2 >> 16) & 0xff) * 3) / 10;

            int c3a = (((c3 >>  0) & 0xff) * 1 + 
                       ((c3 >>  8) & 0xff) * 6 + 
                       ((c3 >> 16) & 0xff) * 3) / 10;

            int c4a = (((c4 >>  0) & 0xff) * 1 + 
                       ((c4 >>  8) & 0xff) * 6 + 
                       ((c4 >> 16) & 0xff) * 3) / 10;

            int c=(((c1a>>4)&0xf)<<12)+
                  (((c2a>>4)&0xf)<<8)+
                  (((c3a>>4)&0xf)<<4)+
                  (((c4a>>4)&0xf));

            aTarget[targetpos] = mAsciiMap[c];                                 
                                 
        }
    }
}
