///////////////////////////////////////////////
// Copyright
///////////////////////////////////////////////
//
// TextFX6
// Copyright (c) 1995-2009 Jari Komppa
//
//
///////////////////////////////////////////////
// License
///////////////////////////////////////////////
// 
//     This software is provided 'as-is', without any express or implied
//     warranty.    In no event will the authors be held liable for any damages
//     arising from the use of this software.
// 
//     Permission is granted to anyone to use this software for any purpose,
//     including commercial applications, and to alter it and redistribute it
//     freely, subject to the following restrictions:
// 
//     1. The origin of this software must not be misrepresented; you must not
//        claim that you wrote the original software. If you use this software
//        in a product, an acknowledgment in the product documentation would be
//        appreciated but is not required.
//     2. Altered source versions must be plainly marked as such, and must not be
//        misrepresented as being the original software.
//     3. This notice may not be removed or altered from any source distribution.
// 
// (eg. same as ZLIB license)
// 
//
///////////////////////////////////////////////

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "textfx.h"


///////////////////////////////////////////////
// Global data & variables
//

#include "tfx_fontdata.h"

///////////////////////////////////////////////
// Functions
//

TFX_ColAsciiArt::TFX_ColAsciiArt()
{
    int i;
    for (i = 0; i < 16; i++)
        mAsciiMap[i] = NULL;
    mOption = ~0;
    mRoll = 0;
}

TFX_ColAsciiArt::~TFX_ColAsciiArt()
{
    //int i;
    //for (i = 0; i < 16; i++)
    //    delete[] mAsciiMap[i];
}



#define IS_BIT(a,b) (((a)&(1<<(7-(b))))?1:0)


// This is _old_ code, and I don't feel like rewriting it at the moment.

void TFX_ColAsciiArt::BuildLUT()
{
    int startchar=0;
    int endchar=255;
    startchar=32;
    endchar=126;
    int a,b,c,d,e,f,g,h,i,m;  
    int reg[256*4];
    int used[256];
    unsigned char * pchar = (unsigned char *)TFX_AsciiFontdata;    
    pchar += startchar * 12;
    for (m = 0; m < 16; m++)
        if (mAsciiMap[m] != NULL)       // regenerating? 
            delete[] mAsciiMap[m];
    for (m = 0; m < 16; m++)
    {
        mAsciiMap[m] = new short[65536]; // allocate the look-up table 
        memset(mAsciiMap[m], 0, sizeof(short) * 65536);
    }
    memset(used,0,sizeof(int)*256);
    // Step one: analyse font 
    for (a=startchar;a<(endchar+1);a++) 
    {
        e=0;
        for (b=0;b<6;b++)
            for (c=0;c<4;c++)
                e+=IS_BIT(*(pchar+b),c);
        *(reg+a*4+0)=(e*16)/24;
        e=0;
        for (b=0;b<6;b++)
            for (c=0;c<4;c++)
                e+=IS_BIT(*(pchar+b),c+4);
        *(reg+a*4+1)=(e*16)/24;
        e=0;
        pchar+=6;
        for (b=0;b<6;b++)
            for (c=0;c<4;c++)
                e+=IS_BIT(*(pchar+b),c);
        *(reg+a*4+2)=(e*16)/24;
        e=0;
        for (b=0;b<6;b++)
            for (c=0;c<4;c++)
                e+=IS_BIT(*(pchar+b),c+4);
        *(reg+a*4+3)=(e*16)/24;
        pchar+=6;
    }

    // Step two: build lookup table with the good old nearest-value mess 
    for (a=0;a<16;a++) 
    {
        for (b=0;b<16;b++)
        {
            for (c=0;c<16;c++)
            {
                for (d=0;d<16;d++) 
                {
                    f=0xffff;
                    g=0;
                    for (h=startchar;h<(endchar+1);h++) 
                    {
                        i=abs(*(reg+h*4+0)-a)+abs(*(reg+h*4+1)-b)+abs(*(reg+h*4+2)-c)+abs(*(reg+h*4+3)-d);
                        if (i<f) 
                        {
                            f=i;
                            g=h;
                        }
                    }
                    m = 0;
                    for (h=startchar;h<(endchar+1);h++) 
                    {
                        i=abs(*(reg+h*4+0)-a)+abs(*(reg+h*4+1)-b)+abs(*(reg+h*4+2)-c)+abs(*(reg+h*4+3)-d);
                        if (i < f + 3) 
                        {   
                            *(mAsciiMap[m]+(a<<12)+(b<<8)+(c<<4)+d)=h;
                            m++;
                            if (m == 16) m = 15;
                        }
                    }
                }
            }            
        }
    }

    for (a = 0; a < 65536; a++)
    {
        c = 0;
        for (b = 0; b < 16; b++)
        {
            if (mAsciiMap[b][a] == 0)
            {
                mAsciiMap[b][a] = mAsciiMap[c][a];
                c++;
            }
        }
    }
}


void TFX_ColAsciiArt::Dump1x(int *aSource, TFXQuad &aSrcQuad, int aSrcPitch, int aTgtX0, int aTgtY0, short *aTarget)
{
    assert(mAsciiMap!=NULL);
    int xsize = (aSrcQuad.x1 - aSrcQuad.x0);
    int ysize = (aSrcQuad.y1 - aSrcQuad.y0);
    int sourceyofs = aSrcQuad.y0 * aSrcPitch;
    mRoll++;

    for (int y = 0; y < ysize; y++, sourceyofs += aSrcPitch)
    {
        for (int x = 0, targetpos=(aTgtY0 + y) * TFX_ConsoleWidth + aTgtX0, sourcepos = sourceyofs + aSrcQuad.x0; x < xsize; x++, targetpos++, sourcepos++)
        {
            int c1 = aSource[sourcepos];

            int c1a = (((c1 >>  0) & 0xff) + 
                       ((c1 >>  8) & 0xff) + 
                       ((c1 >> 16) & 0xff)) / 3;

            int c=((c1a>>4)<<12)+
                  ((c1a>>4)<<8)+
                  ((c1a>>4)<<4)+
                  ((c1a>>4));

            int mixcol = aSource[sourcepos];

            if (mOption & CAA_USEDITHER)
                aTarget[targetpos] = mAsciiMap[(x^y^mRoll)&15][c] & 0xff | findcol(mixcol);                                 
            else
                aTarget[targetpos] = mAsciiMap[0][c] & 0xff | findcol(mixcol);                                 
        }
    }
}

static const int TFX_Palette[16*3]={ // 'clean' RGB palette 
       0, 0, 0,  0, 0,32,  0,32, 0,  0,32,32, 32, 0, 0, 32, 0,32, 32,32, 0, 32,32,32,
9999,9999,9999,  0, 0,63,  0,63, 0,  0,63,63, 63, 0, 0, 63, 0,63, 63,63, 0, 63,63,63};

int TFX_ColAsciiArt::findcol(int incol)
{
    int b = (incol >> 18) & 63;
    int g = (incol >> 10) & 63;
    int r = (incol >> 2) & 63;

    int dist = 0xffffff;
    int bg = 0;
    int i;
#define SQUARE(a) (a)*(a)
    if (mOption & CAA_USEBG)
    {
        for (i = 0; i < 16; i++)
        {
            int d = SQUARE(TFX_Palette[i*3+0]-r) +
                    SQUARE(TFX_Palette[i*3+1]-g) +
                    SQUARE(TFX_Palette[i*3+2]-b);

            if (d < dist)
            {
                bg = i;
                dist = d;
            }
        }
    }

    dist = 0xffffff;
    int fg;
    for (i = 0; i < 16; i++)
    {
        int d = SQUARE(TFX_Palette[i*3+0]-r) +
                SQUARE(TFX_Palette[i*3+1]-g) +
                SQUARE(TFX_Palette[i*3+2]-b);

        if (d < dist && i != bg)
        {
            fg = i;
            dist = d;
        }
    }

    return (bg << 12) | (fg << 8);

}

void TFX_ColAsciiArt::Dump2x(int *aSource, TFXQuad &aSrcQuad, int aSrcPitch, int aTgtX0, int aTgtY0, short *aTarget)
{
    assert(mAsciiMap!=NULL);
    int xsize = (aSrcQuad.x1 - aSrcQuad.x0) / 2;
    int ysize = (aSrcQuad.y1 - aSrcQuad.y0) / 2;
    int sourceyofs = aSrcQuad.y0 * aSrcPitch;
    mRoll++;

    for (int y = 0; y < ysize; y++, sourceyofs += aSrcPitch * 2)
    {
        for (int x = 0, targetpos=(aTgtY0 + y) * TFX_ConsoleWidth + aTgtX0, sourcepos = sourceyofs + aSrcQuad.x0; x < xsize; x++, targetpos++, sourcepos += 2)
        {
            int c1 = aSource[sourcepos];
            int c2 = aSource[sourcepos + 1];
            int c3 = aSource[sourcepos + aSrcPitch];
            int c4 = aSource[sourcepos + aSrcPitch + 1];

            int c1a = (((c1 >>  0) & 0xff) + 
                       ((c1 >>  8) & 0xff) + 
                       ((c1 >> 16) & 0xff)) / 3;

            int c2a = (((c2 >>  0) & 0xff) + 
                       ((c2 >>  8) & 0xff) + 
                       ((c2 >> 16) & 0xff)) / 3;

            int c3a = (((c3 >>  0) & 0xff) + 
                       ((c3 >>  8) & 0xff) + 
                       ((c3 >> 16) & 0xff)) / 3;

            int c4a = (((c4 >>  0) & 0xff) + 
                       ((c4 >>  8) & 0xff) + 
                       ((c4 >> 16) & 0xff)) / 3;

            int c=(((c1a>>4)&0xf)<<12)+
                  (((c2a>>4)&0xf)<<8)+
                  (((c3a>>4)&0xf)<<4)+
                  (((c4a>>4)&0xf));

            int mixcol = ((c1 & 0xf0f0f0) +
                          (c2 & 0xf0f0f0) +
                          (c3 & 0xf0f0f0) +
                          (c4 & 0xf0f0f0)) >> 2;

            if (mOption & CAA_USEDITHER)
                aTarget[targetpos] = mAsciiMap[(x^y^mRoll)&15][c] & 0xff | findcol(mixcol);                                 
            else
                aTarget[targetpos] = mAsciiMap[0][c] & 0xff | findcol(mixcol);                                 
        }
    }
}

void TFX_ColAsciiArt::Dump4x(int *aSource, TFXQuad &aSrcQuad, int aSrcPitch, int aTgtX0, int aTgtY0, short *aTarget)
{
    assert(mAsciiMap!=NULL);
    int xsize = (aSrcQuad.x1 - aSrcQuad.x0) / 4;
    int ysize = (aSrcQuad.y1 - aSrcQuad.y0) / 4;
    int sourceyofs = aSrcQuad.y0 * aSrcPitch;
    mRoll++;

    for (int y = 0; y < ysize; y++, sourceyofs += aSrcPitch * 4)
    {
        for (int x = 0, targetpos=(aTgtY0 + y) * TFX_ConsoleWidth + aTgtX0, sourcepos = sourceyofs + aSrcQuad.x0; x < xsize; x++, targetpos++, sourcepos += 4)
        {
            int c1 = (aSource[sourcepos]>>8)&0xff;            
            c1    += (aSource[sourcepos + 1]>>8)&0xff;
            c1    += (aSource[sourcepos + aSrcPitch]>>8)&0xff;
            c1    += (aSource[sourcepos + aSrcPitch + 1]>>8)&0xff;
            c1   >>= 2;

            int c2 = (aSource[2 + sourcepos]>>8)&0xff;            
            c2    += (aSource[2 + sourcepos + 1]>>8)&0xff;
            c2    += (aSource[2 + sourcepos + aSrcPitch]>>8)&0xff;
            c2    += (aSource[2 + sourcepos + aSrcPitch + 1]>>8)&0xff;
            c2   >>= 2;

            int c3 = (aSource[aSrcPitch + sourcepos]>>8)&0xff;            
            c3    += (aSource[aSrcPitch + sourcepos + 1]>>8)&0xff;
            c3    += (aSource[aSrcPitch + sourcepos + aSrcPitch]>>8)&0xff;
            c3    += (aSource[aSrcPitch + sourcepos + aSrcPitch + 1]>>8)&0xff;
            c3   >>= 2;

            int c4 = (aSource[2 + aSrcPitch + sourcepos]>>8)&0xff;            
            c4    += (aSource[2 + aSrcPitch + sourcepos + 1]>>8)&0xff;
            c4    += (aSource[2 + aSrcPitch + sourcepos + aSrcPitch]>>8)&0xff;
            c4    += (aSource[2 + aSrcPitch + sourcepos + aSrcPitch + 1]>>8)&0xff;
            c4   >>= 2;

            int c1a = (((c1 >>  0) & 0xff) + 
                       ((c1 >>  8) & 0xff) + 
                       ((c1 >> 16) & 0xff)) / 3;

            int c2a = (((c2 >>  0) & 0xff) + 
                       ((c2 >>  8) & 0xff) + 
                       ((c2 >> 16) & 0xff)) / 3;

            int c3a = (((c3 >>  0) & 0xff) + 
                       ((c3 >>  8) & 0xff) + 
                       ((c3 >> 16) & 0xff)) / 3;

            int c4a = (((c4 >>  0) & 0xff) + 
                       ((c4 >>  8) & 0xff) + 
                       ((c4 >> 16) & 0xff)) / 3;

            int c=(((c1a>>4)&0xf)<<12)+
                  (((c2a>>4)&0xf)<<8)+
                  (((c3a>>4)&0xf)<<4)+
                  (((c4a>>4)&0xf));

            int redblue = 0;
            int green = 0;
            for (int iny = 0, inyp = sourcepos; iny < 4; iny++, inyp += aSrcPitch)
            {
                for (int inx = 0, inpos = inyp; inx < 4; inx++, inpos++)
                {
                    redblue += aSource[inpos] & 0xff00ff;
                    green   += aSource[inpos] & 0x00ff00;
                }
            }

            int mixcol = ((redblue>>4) & 0xff00ff) + ((green>>4) & 0x00ff00);

            if (mOption & CAA_USEDITHER)
                aTarget[targetpos] = mAsciiMap[(x^y^mRoll)&15][c] & 0xff | findcol(mixcol);                                 
            else
                aTarget[targetpos] = mAsciiMap[0][c] & 0xff | findcol(mixcol);                                 
        }
    }
}
