///////////////////////////////////////////////
// Copyright
///////////////////////////////////////////////
//
// TextFX6
// Copyright (c) 1995-2009 Jari Komppa
//
//
///////////////////////////////////////////////
// License
///////////////////////////////////////////////
// 
//     This software is provided 'as-is', without any express or implied
//     warranty.    In no event will the authors be held liable for any damages
//     arising from the use of this software.
// 
//     Permission is granted to anyone to use this software for any purpose,
//     including commercial applications, and to alter it and redistribute it
//     freely, subject to the following restrictions:
// 
//     1. The origin of this software must not be misrepresented; you must not
//        claim that you wrote the original software. If you use this software
//        in a product, an acknowledgment in the product documentation would be
//        appreciated but is not required.
//     2. Altered source versions must be plainly marked as such, and must not be
//        misrepresented as being the original software.
//     3. This notice may not be removed or altered from any source distribution.
// 
// (eg. same as ZLIB license)
// 
//
///////////////////////////////////////////////

#include <stdlib.h>
#include <assert.h>
#include "textfx.h"


///////////////////////////////////////////////
// Global data & variables
//

#include "tfx_fontdata.h"

static const int TFX_Palette[16*3]={ // 'clean' RGB palette 
0x00,0x00,0x00,  0x00,0x00,0x80,  0x00,0x80,0x00,  0x00,0x80,0x80, 0x80,0x00,0x00, 0x80,0x00,0x80, 0x80,0x80,0x00, 0x80,0x80,0x80,
0xfff,0xfff,0xfff,  0x00,0x00,0xff,  0x00,0xff,0x00,  0x00,0xff,0xff, 0xff,0x00,0x00, 0xff,0x00,0xff, 0xff,0xff,0x00, 0xff,0xff,0xff};

///////////////////////////////////////////////
// Functions
//

TFX_ColAsciiArt2::TFX_ColAsciiArt2()
{
    mAsciiMap=NULL;
}

TFX_ColAsciiArt2::~TFX_ColAsciiArt2()
{
    delete[] mAsciiMap;
}

static int findcol(int r, int g, int b, int except)
{
    int dist = 0xffffff;
    int col = 0;
    int i;
#define SQUARE(a) (a)*(a)
    for (i = 0; i < 16; i++)
    {
        if (i != except)
        {
            int d = SQUARE(TFX_Palette[i*3+2]-r) +
                    SQUARE(TFX_Palette[i*3+1]-g) +
                    SQUARE(TFX_Palette[i*3+0]-b);

            if (d < dist)
            {
                col = i;
                dist = d;
            }
        }
    }

    return col;
}


#define IS_BIT(a,b) (((a)&(1<<(7-(b))))?1:0)

void TFX_ColAsciiArt2::BuildLUT()
{
    int startchar=0;
    int endchar=255;
    startchar=32;
    endchar=126;
    int a,b,c,d,e,f,g,h,i;  
    int * reg;
    reg = new int[256*4]; // allocate memory for font analyse data 
    unsigned char * pchar = (unsigned char *)TFX_AsciiFontdata;    
    if (mAsciiMap != NULL)       // regenerating? 
        delete[] mAsciiMap;
    mAsciiMap = new short[16*16*16*16]; // allocate the look-up table 
    // Step one: analyse font

    int chars;
    for (chars = startchar; chars <= endchar; chars++)
    {
        int charoffset = chars * 12;
        int i, j, c;
        int scratch[8*12];
        for (i = 0, c = 0; i < 12; i++)
        {
            for (j = 0; j < 8; j++, c++)
            {
                int cc = IS_BIT(*(pchar+i+charoffset),j) ? 255 : 0;
                scratch[i*8+j] = cc;
            }
        }

        int blurred[12*8];

        for (i = 0, c = 0; i < 12; i++)
        {
            for (j = 0; j < 8; j++, c++)
            {
                blurred[c] =
                    (
                    scratch[((i+0)%12)*8+((j+0)%8)]*4 +
                    scratch[((i+1)%12)*8+((j+0)%8)]*2*((i==11)?0:1) +
                    scratch[((i+11)%12)*8+((j+0)%8)]*2*((i==0)?0:1) +
                    scratch[((i+0)%12)*8+((j+1)%8)]*2*((j==7)?0:1) +
                    scratch[((i+1)%12)*8+((j+1)%8)]*((i==11 || j==7)?0:1) +
                    scratch[((i+11)%12)*8+((j+1)%8)]*((i==0 || j==7)?0:1) +
                    scratch[((i+0)%12)*8+((j+7)%8)]*2*((j==0)?0:1) +
                    scratch[((i+1)%12)*8+((j+7)%8)]*((i==11 || j==0)?0:1) +
                    scratch[((i+11)%12)*8+((j+7)%8)]*((i==0 || j==0)?0:1)
                    ) / 8;
            }
        }

        e=0;
        for (b=0;b<6;b++)
            for (c=0;c<4;c++)
                e+=blurred[b*8+c];
        *(reg+chars*4+0)=e/24;
        e=0;
        for (b=0;b<6;b++)
            for (c=0;c<4;c++)
                e+=blurred[b*8+c+4];
        *(reg+chars*4+1)=e/24;
        e=0;
        for (b=0;b<6;b++)
            for (c=0;c<4;c++)
                e+=blurred[b*8+c+6*8];
        *(reg+chars*4+2)=e/24;
        e=0;
        for (b=0;b<6;b++)
            for (c=0;c<4;c++)
                e+=blurred[b*8+c+6*8+4];
        *(reg+chars*4+3)=e/24;
    }

    // Step two: build lookup table with the good old nearest-value mess 
    for (a=0;a<16;a++) 
    {
        for (b=0;b<16;b++)
        {
            for (c=0;c<16;c++)
            {
                for (d=0;d<16;d++) 
                {
                    f=0xffffff;
                    g=0;
                    for (h=startchar;h<(endchar+1);h++) 
                    {
#define POT(x) ((x)*(x))
                        i=POT(*(reg+h*4+0)-a*16)+
                          POT(*(reg+h*4+1)-b*16)+
                          POT(*(reg+h*4+2)-c*16)+
                          POT(*(reg+h*4+3)-d*16);
                        if (i<f) 
                        {
                            f=i;
                            g=h;
                        }
                    }
                    *(mAsciiMap+(a<<12)+(b<<8)+(c<<4)+d)=0x0700+g;
                }
            }            
        }
    }
    delete[] reg;   
}


void TFX_ColAsciiArt2::Dump1x(int *aSource, TFXQuad &aSrcQuad, int aSrcPitch, int aTgtX0, int aTgtY0, short *aTarget)
{
    assert(mAsciiMap!=NULL);
    int xsize = (aSrcQuad.x1 - aSrcQuad.x0);
    int ysize = (aSrcQuad.y1 - aSrcQuad.y0);
    int sourceyofs = aSrcQuad.y0 * aSrcPitch;
    TFXQuad nq;
    nq.x0 = 0;
    nq.y0 = 0;
    nq.x1 = xsize * 2;
    nq.y1 = ysize * 2;
    int *temp = new int[xsize * ysize * 2 * 2];

    for (int y = 0; y < ysize; y++, sourceyofs += aSrcPitch)
    {
        for (int x = 0, targetpos=(aTgtY0 + y) * TFX_ConsoleWidth + aTgtX0, sourcepos = sourceyofs + aSrcQuad.x0; x < xsize; x++, targetpos++, sourcepos ++)
        {
            int c1 = aSource[sourcepos];
            
            temp[y*xsize*2*2+x*2+0] = c1;
            temp[y*xsize*2*2+x*2+1] = c1;
            temp[y*xsize*2*2+x*2+0+xsize*2] = c1;
            temp[y*xsize*2*2+x*2+1+xsize*2] = c1;
        }
    }

    Dump2x(temp,nq,xsize*2,aTgtX0,aTgtY0,aTarget);
    delete[] temp;
}


void TFX_ColAsciiArt2::Dump2x(int *aSource, TFXQuad &aSrcQuad, int aSrcPitch, int aTgtX0, int aTgtY0, short *aTarget)
{
    assert(mAsciiMap!=NULL);
    int xsize = (aSrcQuad.x1 - aSrcQuad.x0) / 2;
    int ysize = (aSrcQuad.y1 - aSrcQuad.y0) / 2;
    int sourceyofs = aSrcQuad.y0 * aSrcPitch;

    for (int y = 0; y < ysize; y++, sourceyofs += aSrcPitch * 2)
    {
        for (int x = 0, targetpos=(aTgtY0 + y) * TFX_ConsoleWidth + aTgtX0, sourcepos = sourceyofs + aSrcQuad.x0; x < xsize; x++, targetpos++, sourcepos += 2)
        {
            int c1 = aSource[sourcepos];
            int c2 = aSource[sourcepos + 1];
            int c3 = aSource[sourcepos + aSrcPitch];
            int c4 = aSource[sourcepos + aSrcPitch + 1];

            int avgcolor = ((c1 & 0xfcfcfc) + (c2 & 0xfcfcfc) + (c3 & 0xfcfcfc) + (c4 & 0xfcfcfc)) >> 2;

            int primarycolor = findcol((avgcolor >> 16) & 0xff, 
                                       (avgcolor >> 8) & 0xff, 
                                       (avgcolor >> 0) & 0xff, -1);
            int secondarycolor = findcol((((avgcolor >> 16) & 0xff) - TFX_Palette[primarycolor * 3 + 2])*2, 
                                         (((avgcolor >> 8) & 0xff) - TFX_Palette[primarycolor * 3 + 1])*2, 
                                         (((avgcolor >> 0) & 0xff) - TFX_Palette[primarycolor * 3 + 0])*2, primarycolor);

//#define CAP(a) (((a) < 0)?0:(a))
#define CAP(a) abs(a)*2


            c1 = (CAP(((c1 >> 0) & 0xff) - TFX_Palette[primarycolor*3+0]) << 0) |
                 (CAP(((c1 >> 8) & 0xff) - TFX_Palette[primarycolor*3+1]) << 8) |
                 (CAP(((c1 >> 16) & 0xff) - TFX_Palette[primarycolor*3+2]) << 16);
            c2 = (CAP(((c2 >> 0) & 0xff) - TFX_Palette[primarycolor*3+0]) << 0) |
                 (CAP(((c2 >> 8) & 0xff) - TFX_Palette[primarycolor*3+1]) << 8) |
                 (CAP(((c2 >> 16) & 0xff) - TFX_Palette[primarycolor*3+2]) << 16);
            c3 = (CAP(((c3 >> 0) & 0xff) - TFX_Palette[primarycolor*3+0]) << 0) |
                 (CAP(((c3 >> 8) & 0xff) - TFX_Palette[primarycolor*3+1]) << 8) |
                 (CAP(((c3 >> 16) & 0xff) - TFX_Palette[primarycolor*3+2]) << 16);
            c4 = (CAP(((c4 >> 0) & 0xff) - TFX_Palette[primarycolor*3+0]) << 0) |
                 (CAP(((c4 >> 8) & 0xff) - TFX_Palette[primarycolor*3+1]) << 8) |
                 (CAP(((c4 >> 16) & 0xff) - TFX_Palette[primarycolor*3+2]) << 16);

#define MAX(a,b) ((a)>(b))?(a):(b)
            int c1a = MAX(((c1 >>  0) & 0xff),
                      MAX(((c1 >>  8) & 0xff),
                          ((c1 >> 16) & 0xff)));
            int c2a = MAX(((c2 >>  0) & 0xff),
                      MAX(((c2 >>  8) & 0xff),
                          ((c2 >> 16) & 0xff)));
            int c3a = MAX(((c3 >>  0) & 0xff),
                      MAX(((c3 >>  8) & 0xff),
                          ((c3 >> 16) & 0xff)));
            int c4a = MAX(((c4 >>  0) & 0xff),
                      MAX(((c4 >>  8) & 0xff),
                          ((c4 >> 16) & 0xff)));

            int c=(((c1a>>4)&0xf)<<12)+
                  (((c2a>>4)&0xf)<<8)+
                  (((c3a>>4)&0xf)<<4)+
                  (((c4a>>4)&0xf));

            aTarget[targetpos] = (mAsciiMap[c] & 0xff) | (primarycolor << 12) | (secondarycolor << 8);
        }
    }
}

void TFX_ColAsciiArt2::Dump4x(int *aSource, TFXQuad &aSrcQuad, int aSrcPitch, int aTgtX0, int aTgtY0, short *aTarget)
{
    assert(mAsciiMap!=NULL);
    int xsize = (aSrcQuad.x1 - aSrcQuad.x0) / 2;
    int ysize = (aSrcQuad.y1 - aSrcQuad.y0) / 2;
    int sourceyofs = aSrcQuad.y0 * aSrcPitch;
    TFXQuad nq;
    nq.x0 = 0;
    nq.y0 = 0;
    nq.x1 = xsize;
    nq.y1 = ysize;
    int *temp = new int[xsize * ysize];

    for (int y = 0, c = 0; y < ysize; y++, sourceyofs += aSrcPitch * 2)
    {
        for (int x = 0, targetpos=(aTgtY0 + y) * TFX_ConsoleWidth + aTgtX0, sourcepos = sourceyofs + aSrcQuad.x0; x < xsize; x++, targetpos++, sourcepos += 2, c++)
        {
            int c1 = aSource[sourcepos];
            int c2 = aSource[sourcepos + 1];
            int c3 = aSource[sourcepos + aSrcPitch];
            int c4 = aSource[sourcepos + aSrcPitch + 1];

            int avgcolor = ((c1 & 0xfcfcfc) + (c2 & 0xfcfcfc) + (c3 & 0xfcfcfc) + (c4 & 0xfcfcfc)) >> 2;
            
            temp[c] = avgcolor;
        }
    }

    Dump2x(temp,nq,xsize,aTgtX0,aTgtY0,aTarget);
    delete[] temp;
}
