///////////////////////////////////////////////
// Copyright
///////////////////////////////////////////////
//
// TextFX6
// Copyright (c) 1995-2008 Jari Komppa
//
//
///////////////////////////////////////////////
// License
///////////////////////////////////////////////
// 
//     This software is provided 'as-is', without any express or implied
//     warranty.    In no event will the authors be held liable for any damages
//     arising from the use of this software.
// 
//     Permission is granted to anyone to use this software for any purpose,
//     including commercial applications, and to alter it and redistribute it
//     freely, subject to the following restrictions:
// 
//     1. The origin of this software must not be misrepresented; you must not
//        claim that you wrote the original software. If you use this software
//        in a product, an acknowledgment in the product documentation would be
//        appreciated but is not required.
//     2. Altered source versions must be plainly marked as such, and must not be
//        misrepresented as being the original software.
//     3. This notice may not be removed or altered from any source distribution.
// 
// (eg. same as ZLIB license)
// 
//
///////////////////////////////////////////////

#include <stdlib.h>
#include <assert.h>
#include "textfx.h"





///////////////////////////////////////////////
// Compilation options
//


///////////////////////////////////////////////
// Global data & variables
//

// The 9999,9999,9999 is the 'light black' color.
// For some reason NT console in fullscreen shows this color
// as color 7 (42,42,42) instead of the correct color (21,21,21)
// Thus, it is better not used here. 
// Value 9999 moves it far enough from the rgb cube.
// ('clean' palette version would be 32,32,32)

static const int TFX_Palette[16*3]={ // 'clean' RGB palette 
       0, 0, 0,  0, 0,32,  0,32, 0,  0,32,32, 32, 0, 0, 32, 0,32, 32,32, 0, 32,32,32,
9999,9999,9999,  0, 0,63,  0,63, 0,  0,63,63, 63, 0, 0, 63, 0,63, 63,63, 0, 63,63,63};


///////////////////////////////////////////////
// Functions
//


TFX_HalfBlockColorWide::TFX_HalfBlockColorWide()
{
    mOption = 0;
}

TFX_HalfBlockColorWide::~TFX_HalfBlockColorWide()
{
}


#define exp2(a) ((a)*(a))

short int TFX_HalfBlockColorWide::CalcColor(int red, int green, int blue) 
{
    int c,d,ch,co;
    int lastdist, dist;
  
    lastdist = 0x7fffffff;

    for (c = 0, d = 0; c < 16; c++, d += 3) 
    {
        dist = exp2(TFX_Palette[d + 0] - red) +
               exp2(TFX_Palette[d + 1] - green) +
               exp2(TFX_Palette[d + 2] - blue);
        if (dist < lastdist) 
        {
            lastdist = dist;
            co = c;
            ch = 32;
        }
    }
    co = co | (co << 4);
    return((co<<8) + ch);
}



void TFX_HalfBlockColorWide::BuildLUT() 
{
}



void TFX_HalfBlockColorWide::Dump1x(int *aSource, TFXQuad &aSrcQuad, int aSrcPitch, int aTgtX0, int aTgtY0, short *aTarget)
{
    int xsize = aSrcQuad.x1 - aSrcQuad.x0;
    int ysize = aSrcQuad.y1 - aSrcQuad.y0;
    int sourceyofs = aSrcQuad.y0 * aSrcPitch;

    for (int y = 0; y < ysize; y++, sourceyofs+=aSrcPitch)
    {
        for (int x = 0, targetpos = (aTgtY0 + y) * TFX_ConsoleWidth + aTgtX0, sourcepos = sourceyofs + aSrcQuad.x0; x < xsize; x++, targetpos++, sourcepos++)
        {
            int color = aSource[sourcepos];
            aTarget[targetpos] = CalcColor(((color    ) & 0xff)>>2,
                                           ((color>> 8) & 0xff)>>2,
                                           ((color>>16) & 0xff)>>2);
        }
    }
}

int TFX_HalfBlockColorWide::findcol(int incol)
{
    int b = (incol >> 18) & 63;
    int g = (incol >> 10) & 63;
    int r = (incol >> 2) & 63;

    static int lasterr_r = 0;
    static int lasterr_g = 0;
    static int lasterr_b = 0;

    int dist = 0xffffff;
    int bg = 0;
    int i;
#define SQUARE(a) (a)*(a)
    for (i = 0; i < 16; i++)
    {
        int d = SQUARE(TFX_Palette[i*3+0]-r) +
                SQUARE(TFX_Palette[i*3+1]-g) +
                SQUARE(TFX_Palette[i*3+2]-b);

        if (d < dist)
        {
            bg = i;
            dist = d;
            lasterr_r = TFX_Palette[i*3+0]-r;
            lasterr_g = TFX_Palette[i*3+1]-g;
            lasterr_b = TFX_Palette[i*3+2]-b;
        }
    }
    return bg;
}


void TFX_HalfBlockColorWide::Dump2x(int *aSource, TFXQuad &aSrcQuad, int aSrcPitch, int aTgtX0, int aTgtY0, short *aTarget)
{
    int xsize = (aSrcQuad.x1 - aSrcQuad.x0) / 2;
    int ysize = (aSrcQuad.y1 - aSrcQuad.y0) / 2;
    int sourceyofs = aSrcQuad.y0 * aSrcPitch;

    for (int y = 0; y < ysize; y++, sourceyofs += aSrcPitch * 2)
    {
        for (int x = 0, targetpos=(aTgtY0 + y) * TFX_ConsoleWidth + aTgtX0, sourcepos = sourceyofs + aSrcQuad.x0; x < xsize; x++, targetpos++, sourcepos += 2)
        {
            int redblue;
            int green;
            redblue  = aSource[sourcepos] & 0xff00ff;
            green    = aSource[sourcepos] & 0x00ff00;
            redblue += aSource[sourcepos + aSrcPitch] & 0xff00ff;
            green   += aSource[sourcepos + aSrcPitch] & 0x00ff00;
            int color1 = ((redblue>>1) & 0xff00ff) + ((green>>1) & 0x00ff00);
            redblue = green = 0;
            redblue += aSource[sourcepos + 1] & 0xff00ff;
            green   += aSource[sourcepos + 1] & 0x00ff00;
            redblue += aSource[sourcepos + aSrcPitch + 1] & 0xff00ff;
            green   += aSource[sourcepos + aSrcPitch + 1] & 0x00ff00;
            int color2 = ((redblue>>1) & 0xff00ff) + ((green>>1) & 0x00ff00);

            aTarget[targetpos] = 221 | (findcol(color1) << 8) | (findcol(color2) << 12);
        }
    }
}



void TFX_HalfBlockColorWide::Dump4x(int *aSource, TFXQuad &aSrcQuad, int aSrcPitch, int aTgtX0, int aTgtY0, short *aTarget)
{
    int xsize = (aSrcQuad.x1 - aSrcQuad.x0) / 4;
    int ysize = (aSrcQuad.y1 - aSrcQuad.y0) / 4;
    int sourceyofs = aSrcQuad.y0 * aSrcPitch;

    for (int y = 0; y < ysize; y++, sourceyofs += aSrcPitch * 4)
    {
        for (int x = 0, targetpos = (aTgtY0 + y) * TFX_ConsoleWidth + aTgtX0, sourcepos = sourceyofs + aSrcQuad.x0; x < xsize; x++, targetpos++, sourcepos += 4)
        {
            int redblue = 0;
            int green = 0;

            for (int iny = 0, inyp = sourcepos; iny < 4; iny++, inyp += aSrcPitch)
            {
                for (int inx = 0, inpos = inyp; inx < 2; inx++, inpos++)
                {
                    redblue += aSource[inpos] & 0xff00ff;
                    green   += aSource[inpos] & 0x00ff00;
                }
            }

            int color1 = ((redblue>>3) & 0xff00ff) + ((green>>3) & 0x00ff00);

            redblue = 0;
            green = 0;

            for (int iny = 0, inyp = sourcepos; iny < 4; iny++, inyp += aSrcPitch)
            {
                for (int inx = 0, inpos = inyp+2; inx < 2; inx++, inpos++)
                {
                    redblue += aSource[inpos] & 0xff00ff;
                    green   += aSource[inpos] & 0x00ff00;
                }
            }

            int color2 = ((redblue>>3) & 0xff00ff) + ((green>>3) & 0x00ff00);

            aTarget[targetpos] = 221 | (findcol(color1) << 8) | (findcol(color2) << 12);
        }
    }
}
